#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <nllibc.h>

static FILE __nlstdin;
static FILE __nlstdout;
static FILE __nlstderr;

FILE *stdin  = &__nlstdin;
FILE *stdout = &__nlstdout;
FILE *stderr = &__nlstderr;

static char *_utoa(unsigned long val)
{
  static char s[32];
  char *p;

  p = s + sizeof(s);
  *(--p) = '\0';
  while (val) {
    *(--p) = "0123456789"[val % 10];
    val /= 10;
  }
  if (*p == '\0')
    *(--p) = '0';

  return p;
}

static char *_itoa(long val)
{
  char *p;

  p = _utoa((val < 0) ? -val : val);
  if (val < 0)
    *(--p) = '-';

  return p;
}

static char *_otoa(unsigned long val)
{
  static char s[32];
  char *p;

  p = s + sizeof(s);
  *(--p) = '\0';
  while (val) {
    *(--p) = "01234567"[val & 0x7];
    val >>= 3;
  }
  if (*p == '\0')
    *(--p) = '0';

  return p;
}

static char *_xtoa(unsigned long val, int upper)
{
  static char s[32];
  char *p;
  char *x;

  x = upper ? "0123456789ABCDEF" : "0123456789abcdef";

  p = s + sizeof(s);
  *(--p) = '\0';
  while (val) {
    *(--p) = x[val & 0xf];
    val >>= 4;
  }
  if (*p == '\0')
    *(--p) = '0';

  return p;
}

static char *outc(char *p, int fd, int c)
{
  unsigned char ch = c;
  if (p) {
    *(p++) = c;
    *p = '\0';
  }
  if (fd >= 0) {
    write(fd, &ch, 1);
  }
  return p;
}

static char *outs(char *p, int fd, char *s)
{
  int len = strlen(s);
  if (p) {
    strcpy(p, s);
    p += len;
  }
  if (fd >= 0) {
    write(fd, s, len);
  }
  return p;
}

static int va_sprintf(char *str, int fd, const char *format, va_list ap)
{
  char *p = str, *sv;
  const char *f;
  enum {
    MODE_NORMAL = 0,
    MODE_INT = 1,
    MODE_LONG = 2,
    MODE_QUAD = 3,
  } mode;

  mode = MODE_NORMAL;

  if (p)
    *p = '\0';

  for  (f = format; *f; f++) {
    if (mode != MODE_NORMAL) {
      switch (*f) {
      case 'c':
	p = outc(p, fd, va_arg(ap, int));
	break;
      case 's':
	p = outs(p, fd, va_arg(ap, char *));
	break;
      case 'l':
	switch (mode) {
	case MODE_INT:  mode = MODE_LONG; break;
	case MODE_LONG: mode = MODE_QUAD; break;
	default: break;
	}
	break;
      case 'q':
	switch (mode) {
	case MODE_INT:
	case MODE_LONG: mode = MODE_QUAD; break;
	default: break;
	}
	break;
      case 'd':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = _itoa(va_arg(ap, long)); break;
	case MODE_INT:
	default:        sv = _itoa(va_arg(ap, int)); break;
	}
	p = outs(p, fd, sv);
	break;
      case 'u':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = _utoa(va_arg(ap, unsigned long));
	case MODE_INT:
	default:        sv = _utoa(va_arg(ap, unsigned int));
	}
	p = outs(p, fd, sv);
	break;
      case 'o':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = _otoa(va_arg(ap, unsigned long));
	case MODE_INT:
	default:        sv = _otoa(va_arg(ap, unsigned int));
	}
	p = outs(p, fd, sv);
	break;
      case 'x':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = _xtoa(va_arg(ap, unsigned long), 0);
	case MODE_INT:
	default:        sv = _xtoa(va_arg(ap, unsigned int), 0);
	}
	p = outs(p, fd, sv);
	break;
      case 'X':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = _xtoa(va_arg(ap, unsigned long), 1);
	case MODE_INT:
	default:        sv = _xtoa(va_arg(ap, unsigned int), 1);
	}
	p = outs(p, fd, sv);
	break;
      case 'p':
	p = outs(p, fd, "0x");
	p = outs(p, fd, _xtoa((unsigned long)va_arg(ap, char *), 0));
	break;
      case '%':
	p = outc(p, fd, '%');
	break;
      default:
	break;
      }
      mode = MODE_NORMAL;
      continue;
    }

    if (*f == '%') {
      mode = MODE_INT;
      continue;
    }

    p = outc(p, fd, *f);
  }

  return 0;
}

int sprintf(char *str, const char *format, ...)
{
  va_list arg;
  int done;

  va_start(arg, format);
  done = va_sprintf(str, -1, format, arg);
  va_end(arg);

  return done;
}

int printf(const char *format, ...)
{
  va_list arg;
  int done;

  va_start(arg, format);
  done = va_sprintf(NULL, 1, format, arg);
  va_end(arg);

  return done;
}

int puts(const char *str)
{
  write(1, str, strlen(str));
  write(1, "\n", 1);
  return 0;
}

char *fgets(char *str, int size, FILE *stream)
{
  int r;
  r = read(0, str, size);
  if (r <= 0)
    return NULL;
  return str;
}

int fputs(const char *str, FILE *stream)
{
  return write(1, str, strlen(str));
}
